def countdown(num: int, step: int):
    while num > 0:
        if num > step:
            yield step
        else:
            yield num
        num -= step


def dictfilter(d=None, **kw):
    """Remove all keys from dict ``d`` whose value is :const:`None`."""
    d = kw if d is None else (dict(d, **kw) if kw else d)
    return {k: v for k, v in d.items() if v is not None}



import argparse
import logging
import sys
import asyncio
from samping import run_worker as run

from .app import app


def run_worker(queues: str, num_worker: int):
    root = logging.getLogger()
    root.setLevel(logging.INFO)

    logging.getLogger("samping").setLevel(logging.DEBUG)
    logging.getLogger("example").setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    root.addHandler(handler)

    asyncio.run(run(app, queues, num_worker, beat=True))


if __name__ == "__main__":
    parser = argparse.ArgumentParser("serpengine worker")
    parser.add_argument("--queue", help="queue to process", default="samping")
    parser.add_argument("--num-worker", type=int, default=3)

    args = parser.parse_args()

    run_worker(args.queue, args.num_worker)



import asyncio
import logging
import signal
from datetime import datetime, timezone

from .driver import QueueDriver
from .types import ASGI3Application
from .messages import Message
from .channel import channel, Sender
from .protocol import LifeSpanCyle, QueueCycle, beat_cycle
from .utils.timer import Timer
from .utils.time import get_exponential_backoff_interval
from .tasks import TaskStatus

logger = logging.getLogger("samping")

__all__ = [
    "Worker",
    "run_worker",
]


async def run_worker(app, queues: str, num_worker: int = 3, beat=False):
    worker = Worker(app, driver=app.driver)
    await worker.start(queues, num_worker=num_worker, beat=beat)


async def wrap_future(e, f):
    return e, await f


async def select(*futures):
    done, pendings = await asyncio.wait(
        [asyncio.ensure_future(wrap_future(label, fut)) for label, fut in futures],
        return_when=asyncio.FIRST_COMPLETED,
    )
    for task in pendings:
        if not task.cancelled():
            task.cancel()

    return done


class Worker:
    def __init__(self, app: ASGI3Application, driver: QueueDriver):
        self._app = app
        self._driver = driver
        self._stopping = asyncio.Event()

    def signal_handler(self, signame):
        logger.info("receive signal %s! exiting", signame)
        self._stopping.set()

    def install_signal_handler(self):
        loop = asyncio.get_running_loop()

        for signame in ("SIGINT", "SIGTERM"):
            loop.add_signal_handler(
                getattr(signal, signame),
                lambda: self.signal_handler(signame),
            )

    async def start(self, queues: str, num_worker: int = 3, beat=False):
        logger.info("starting %d workers processing queue %s", num_worker, queues)

        self.install_signal_handler()

        async with LifeSpanCyle(self._app) as lifespan:
            if lifespan.should_exit:
                return

            timer = Timer(60)
            if beat:
                timer.start()

            msg_tx, msg_rx = channel()
            producer = asyncio.create_task(self._run_producer(msg_tx, queues))
            workers = set()
            semaphore = asyncio.Semaphore(num_worker)
            ev_tx, ev_rx = channel()

            pending_tasks = 0
            while True:
                selected = await select(
                    ("message", msg_rx.recv()),
                    ("ending", self._stopping.wait()),
                    ("beat", timer.chan()),
                    ("task_event", ev_rx.recv()),
                )

                for task in selected:
                    which, result = task.result()
                    if which == "message":
                        logger.debug(
                            "receive message from broker, current pending task: %d",
                            pending_tasks,
                        )
                        task = asyncio.create_task(
                            self.on_message_received(semaphore, ev_tx, result)
                        )
                        workers.add(task)
                        task.add_done_callback(workers.discard)
                    elif which == "ending":
                        logger.info("warm shutdown..")
                        break
                    elif which == "beat":
                        logger.debug("receive beat event")
                        task = asyncio.create_task(
                            beat_cycle(self._app, datetime.now(timezone.utc))
                        )
                        workers.add(task)
                        task.add_done_callback(workers.discard)
                    elif which == "task_event":
                        logger.debug("task event received: %s", result)
                        if result == TaskStatus.PENDING:
                            pending_tasks += 1
                        elif result == TaskStatus.FINISHED:
                            pending_tasks -= 1
                else:
                    if producer.done():
                        logger.info("queue producer exited, starting producer")
                        producer = asyncio.create_task(self._run_producer(queues))
                    continue

                break

            if pending_tasks > 0:
                while pending_tasks > 0:
                    selected = await select(
                        ("ending", self._stopping.wait()), ("task_event", ev_rx.recv())
                    )
                    for task in selected:
                        which, result = task.result()
                        if which == "ending":
                            logger.info("Okay fine, shutting down now. See ya!")
                            break
                        elif which == "task_event":
                            if result == TaskStatus.PENDING:
                                pending_tasks += 1
                            elif result == TaskStatus.FINISHED:
                                pending_tasks -= 1
                    else:
                        continue

                    break

    async def on_message_received(
        self, semaphore: asyncio.Semaphore, tx: Sender[TaskStatus], message: Message
    ):
        async with semaphore:
            queue = QueueCycle(message, driver=self._driver, sender=tx)
            await queue.run_app(self._app)

    async def _run_producer(self, msg_tx: Sender[Message], queues: str):
        retries = 0
        while not self._stopping.is_set():
            try:
                async for message in self._driver.consume(queues):
                    await msg_tx.send(message)
            except BaseException:
                if not self._stopping.is_set():
                    retries += 1
                    wait = get_exponential_backoff_interval(2, retries, maximum=60)
                    logger.error(
                        "producer exited, waiting %d seconds to restart it",
                        wait,
                        exc_info=True,
                    )
                    await asyncio.sleep(wait)
                else:
                    break



import asyncio
import logging
from datetime import datetime, timezone, timedelta
from croniter import croniter
from functools import cached_property
import traceback
import uuid
import numbers
from typing import Optional, List, Callable, Union, Dict

from ..driver import QueueDriver
from ..tasks import Task, CronJob
from ..messages import Message
from ..exceptions import DecodeError, ContentDisallowed
from ..routes import route, Rule
from ..utils.time import (
    timezone as utils_timezone,
    to_utc,
    maybe_make_aware,
    maybe_iso8601,
    utcnow,
)
from ..types import Lifespan, AppType, Scope, Receive, Send
from .conf import Conf
from .tracer import build_tracer, Tracer


async def run_tab(logger, tab, name):
    logger.info("executing cron tab: %s", name)
    await tab.run()


async def wrap_future(e, f):
    return e, await f


async def select(*futures):
    done, pendings = await asyncio.wait(
        [asyncio.ensure_future(wrap_future(label, fut)) for label, fut in futures],
        return_when=asyncio.FIRST_COMPLETED,
    )
    for task in pendings:
        if not task.cancelled():
            task.cancel()

    return done


default_conf = {
    "timezone": None,
    "task_serializer": "msgpack",
}


class _DefaultLifespan:
    async def __aenter__(self):
        pass

    async def __aexit__(self, *exc_info: object) -> None:
        pass

    def __call__(self, app) -> None:
        return self


class App:
    def __init__(
        self,
        driver_factory: Callable[[], QueueDriver],
        default_queue: str = "default",
        routes: Optional[List[Rule]] = None,
        lifespan: Lifespan[AppType] | None = None,
    ):
        self.driver_factory = driver_factory
        self._driver: Union[None, QueueDriver] = None
        self.logger = logging.getLogger("samping")
        self._tasks: Dict[str, Task] = {}
        self._cron_tabs: Dict[str, CronJob] = {}
        self.default_queue = default_queue
        self.routes = routes or None
        self.lifespan_context = lifespan if lifespan else _DefaultLifespan()
        self._conf = Conf(default_conf)

    def task_route(self, task: Task) -> str:
        queue = route(task.name, self.routes)
        self.logger.debug(
            "route %s to queue %s", task.name, queue or self.default_queue
        )
        return queue or self.default_queue

    @property
    def driver(self):
        if self._driver is None:
            self._driver = self.driver_factory()

        return self._driver

    def now(self):
        now_in_utc = to_utc(utcnow())
        return now_in_utc.astimezone(self.timezone)

    def tab(self, **opts):
        def create_job_cls(fun):
            return self._cron_from_fun(fun, **opts)

        return create_job_cls

    def task(self, **opts):
        def create_task_cls(fun):
            return self._task_from_fun(fun, **opts)

        return create_task_cls

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        assert scope["type"] in ("lifespan", "beat", "queue")

        if scope["type"] == "lifespan":
            return await self.lifespan(scope, receive, send)
        if scope["type"] == "beat":
            return await self.beat(scope)

        headers = dict(scope["headers"])
        task_name = headers.get("task", None)
        if task_name:
            return await self.handle_v2_message(scope, receive, send)
        else:
            return await self.handle_v1_message(scope, receive, send)

    async def beat(self, scope: Scope):
        try:
            event_time = maybe_iso8601(scope.get("time", ""))
            event_time = event_time.astimezone(tz=self.timezone)
        except ValueError:
            event_time = self.now()

        await self.run_matched_tabs(current_date=event_time)

    async def handle_v1_message(self, scope: Scope, receive: Receive, send: Send):
        headers = dict(scope["headers"])
        message = Message(
            scope["body"],
            headers=headers,
            properties=dict(scope.get("properties", {}) or {}),
            content_type=headers.get("content_type", None),
            content_encoding=headers.get("content_encoding", None),
        )
        try:
            task_message = message.decode()
        except (DecodeError, ContentDisallowed):
            self.logger.warning("Received invalid message, discarding")
            return await send({"type": "queue.ack", "task": task_name})
        else:
            task_name = task_message["task"]
            if task_name not in self._tasks:
                self.logger.warning(
                    "Receive message for task %s, but it is not in registered task, discarding.",
                    task_name,
                )
                await send({"type": "queue.ack", "task": task_name})
                return

            task = self._tasks[task_name]

            tracer = build_tracer(self, task, message, scope.get("hostname", ""))

            await self.trace_task_execution(scope, send, tracer)

    async def handle_v2_message(self, scope: Scope, receive: Receive, send: Send):
        headers = dict(scope["headers"])
        task_name = headers.get("task", None)

        if task_name not in self._tasks:
            self.logger.warning(
                "Receive message for task %s, but it is not in registered task, discarding.",
                task_name,
            )
            await send({"type": "queue.ack", "task": task_name})
            return

        task = self._tasks[task_name]
        message = Message(
            scope["body"],
            headers=headers,
            properties=dict(scope.get("properties", {}) or {}),
            content_type=headers.get("content_type", None),
            content_encoding=headers.get("content_encoding", None),
        )
        try:
            tracer = build_tracer(self, task, message, scope.get("hostname", ""))
        except (DecodeError, ContentDisallowed):
            self.logger.warning("Received invalid message, discarding")
            return await send({"type": "queue.ack", "task": task_name})

        await self.trace_task_execution(scope, send, tracer)

    async def trace_task_execution(self, scope: Scope, send: Send, tracer: Tracer):
        if tracer.is_delayed:
            await tracer.wait()

        if not tracer.acks_late:
            await send({"type": "queue.ack", "task": tracer.task.name})

        eta = await tracer.trace()
        if eta:
            if isinstance(eta, numbers.Real):
                now = self.now()
                eta = maybe_make_aware(now + timedelta(seconds=eta), tz=self.timezone)

            await self.driver.send_batch(scope["name"], [tracer.as_retry_message(eta)])
        else:
            if tracer.acks_late:
                await send({"type": "queue.ack", "task": tracer.task.name})

    async def lifespan(self, scope: Scope, receive: Receive, send: Send):
        started = False
        await receive()
        try:
            async with self.lifespan_context(self) as maybe_state:
                if maybe_state is not None:
                    if "state" not in scope:
                        raise RuntimeError(
                            "The server/worker does not support state in the lifespan scope"
                        )
                    scope["state"].update(maybe_state)
                await send({"type": "lifespan.startup.complete"})
                started = True
                await receive()
        except BaseException:
            exc_text = traceback.format_exc()
            if started:
                await send({"type": "lifespan.shutdown.failed", "message": exc_text})
            else:
                await send({"type": "lifespan.startup.failed", "message": exc_text})
        else:
            await send({"type": "lifespan.shutdown.complete"})

    async def run_matched_tabs(self, current_date: Optional[datetime] = None):
        n = self.now() if current_date is None else current_date
        tasks = []
        for name, tab in self._cron_tabs.items():
            if croniter.match(tab.expression, n):
                task = asyncio.create_task(
                    run_tab(self.logger, tab, name),
                    name=name + "-" + str(uuid.uuid4()),
                )
                tasks.append(task)
        if tasks:
            try:
                done, _ = await asyncio.wait(tasks)
                for task in done:
                    try:
                        await task
                    except Exception as e:
                        self.logger.error(
                            f"cron job {task.get_name()} exited with error: {e}"
                        )
            except Exception as e:
                self.logger.error(f"cron jobs exited with error: {e}")

    async def _send_message(
        self, message: Message, queue: Optional[str] = None, **kwargs
    ):
        return await self.send_batch(
            messages=[message], queue=queue or self.default_queue, **kwargs
        )

    async def send_batch(
        self, messages: List[Message], queue: Optional[str] = None, **kwargs
    ):
        return await self.driver.send_batch(
            queue=queue or self.default_queue, messages=messages, **kwargs
        )

    def _task_from_fun(self, fun, base=None, name=None, bind=False, **options) -> Task:
        name = name or self.gen_task_name(fun.__name__, fun.__module__)
        base = base or Task
        if name not in self._tasks:
            run = fun if bind else staticmethod(fun)
            task = type(
                fun.__name__,
                (base,),
                dict(
                    {
                        "app": self,
                        "name": name,
                        "run": run,
                        "_decorated": True,
                        "__doc__": fun.__doc__,
                        "__module__": fun.__module__,
                        "__annotations__": fun.__annotations__,
                        "__wrapped__": fun,
                    },
                    **options,
                ),
            )()
            try:
                task.__qualname__ = fun.__qualname__
            except AttributeError:
                pass

            self._tasks[task.name] = task
            task.bind(self)
        else:
            task = self._tasks[name]

        return task

    def _cron_from_fun(
        self, fun, base=None, name=None, bind=False, **options
    ) -> CronJob:
        name = name or self.gen_task_name(fun.__name__, fun.__module__)
        base = base or CronJob
        if name not in self._cron_tabs:
            run = fun if bind else staticmethod(fun)
            task = type(
                fun.__name__,
                (base,),
                dict(
                    {
                        "app": self,
                        "name": name,
                        "run": run,
                        "_decorated": True,
                        "__doc__": fun.__doc__,
                        "__module__": fun.__module__,
                        "__annotations__": fun.__annotations__,
                        "__wrapped__": fun,
                    },
                    **options,
                ),
            )()
            try:
                task.__qualname__ = fun.__qualname__
            except AttributeError:
                pass
            self._cron_tabs[task.name] = task
        else:
            task = self._cron_tabs[name]
        return task

    def gen_task_name(self, name: str, module_name: str) -> str:
        return ".".join(p for p in (module_name, name) if p)

    @property
    def conf(self):
        """Current configuration"""
        return self._conf

    @cached_property
    def timezone(self):
        conf = self.conf
        if not conf.timezone:
            return timezone.utc

        return utils_timezone.get_timezone(conf.timezone)



from types import TracebackType
from typing import TypeVar


BaseExceptionType = TypeVar("BaseException", bound=BaseException)


def reraise(
    tp: type[BaseExceptionType],
    value: BaseExceptionType,
    tb: TracebackType | None = None,
):
    if value.__traceback__ is not tb:
        raise value.with_traceback(tb)
    raise value


class SampingError(BaseException):
    "Base error for Samping"


class TaskPredicate(SampingError):
    """Base class for task-related semi-predicates."""


class Retry(TaskPredicate):
    #: Optional message
    message = None

    #: Exception (if any) that caused the retry to happen.
    exc = None

    #: Time of retry (ETA) in seconds
    when = None

    def __init__(self, exc=None, when=None, **kwargs):
        self.when = when
        super().__init__(self, exc, when, **kwargs)


class TaskExpectedError(TaskPredicate):
    """An error that is expected to happen every once in a while"""


class TaskTimeout(TaskPredicate):
    """Raised when a task runs over its time limit specified by the
    `time_limit` setting
    """


class TraceError(SampingError):
    """Errors that can occur while tracing a task"""


class TaskExpirationError(TraceError):
    """Raised when an expired task is received"""


class OperationalError(SampingError):
    """Recoverable message transport connection error."""


class SerializationError(SampingError):
    """Failed to serialize/deserialize content."""


class EncodeError(SerializationError):
    """Cannot encode object."""


class DecodeError(SerializationError):
    """Cannot decode object."""


class NotBoundError(SampingError):
    """Trying to call channel dependent method on unbound entity."""


class MessageStateError(SampingError):
    """The message has already been acknowledged."""


class LimitExceeded(SampingError):
    """Limit exceeded."""


class ConnectionLimitExceeded(LimitExceeded):
    """Maximum number of simultaneous connections exceeded."""


class ChannelLimitExceeded(LimitExceeded):
    """Maximum number of simultaneous channels exceeded."""


class VersionMismatch(SampingError):
    """Library dependency version mismatch."""


class SerializerNotInstalled(SampingError):
    """Support for the requested serialization type is not installed."""


class ContentDisallowed(SerializerNotInstalled):
    """Consumer does not allow this content-type."""


class InconsistencyError(ConnectionError):
    """Data or environment has been found to be inconsistent.

    Depending on the cause it may be possible to retry the operation.
    """



import logging
import typing
from datetime import datetime

from .lifespan import LifeSpanCyle
from .queue import QueueCycle
from ..types import ASGI3Application


__all__ = [
    "LifeSpanCyle",
    "QueueCycle",
    "beat_cycle",
]

logger = logging.getLogger("samping")


async def default_receive():
    pass


async def default_send(message):
    logger.warning("asgi send called when it shouldn't")


async def beat_cycle(app: ASGI3Application, time: typing.Union[str, datetime]):
    try:
        scope = {
            "type": "beat",
            "time": time if isinstance(time, str) else time.isoformat(),
        }
        await app(scope, default_receive, default_send)
    except BaseException as exc:
        msg = "Exception in Samping application\n"
        logger.error(msg, exc_info=exc)



from samping import LambdaAdapter
from samping.driver.sqs import SQSDriver
import uuid

from .example_app import app, test_task, test_task_v1


def test_execute_sqs_event(mocker):
    driver = SQSDriver(
        endpoint_url="http://localhost:9324",
        use_ssl=False,
        prefetch_size=15,
        visibility_timeout=60,
    )
    message = test_task_v1.to_message(args=[10])
    message_v2 = test_task.to_message(args=["lambda-test"])
    sqs_message1 = driver.message_to_sqs(message)
    sqs_message2 = driver.message_to_sqs(message_v2)

    spy = mocker.spy(app.driver, "send_ack")

    # lambda message event
    event = {
        "Records": [
            {
                "messageId": uuid.uuid4().hex,
                "receiptHandle": uuid.uuid4().hex,
                "attributes": {
                    "ApproximateReceiveCount": "2",
                },
                "body": m["MessageBody"],
                "eventSource": "aws:sqs",
                "eventSourceARN": "arn:aws:sqs:us-east-2:123456789012:samping",
                "awsRegion": "us-east-2",
            }
            for m in [sqs_message1, sqs_message2]
        ]
    }
    lambda_handler = LambdaAdapter(app)

    lambda_handler(event, {})

    assert spy.call_count == 2


def test_execute_event_bridget_schedule(mocker):
    # lambda eventbridget event structure
    event = {
        "version": "0",
        "id": "53dc4d37-cffa-4f76-80c9-8b7d4a4d2eaa",
        "detail-type": "Scheduled Event",
        "source": "aws.events",
        "account": "123456789012",
        "time": "2024-11-08T16:53:06Z",
        "region": "us-east-1",
        "resources": ["arn:aws:events:us-east-1:123456789012:rule/my-scheduled-rule"],
        "detail": {},
    }

    spy = mocker.spy(app, "beat")

    lambda_handler = LambdaAdapter(app)
    lambda_handler(event, {})

    assert spy.call_count == 1



import asyncio
import logging
import time

from ..tasks import Task, Request
from ..channel import Sender
from ..messages import Message
from ..exceptions import TaskExpectedError, TaskExpirationError, TaskTimeout, Retry

logger = logging.getLogger("samping")


class Tracer:
    def __init__(self, task: Task):
        self.task = task

    async def trace(self, monotonic=time.monotonic):
        """Trace the execution of task, return None if succeed, otherwise
        return datetime or seconds when this task should be retried
        """
        if self.is_expired:
            logger.warning(
                "Task %s expired, discarding %s", self.task.name, self.task.request.id
            )
            raise TaskExpirationError()

        start = monotonic()
        try:
            result = await self.execute_task()
        except (TaskTimeout, asyncio.exceptions.TimeoutError) as e:
            duration = monotonic() - start
            logger.warning(
                "Task %d timed out after %fs",
                self.task.name,
                duration,
            )
            return await self.handle_error(e, True)
        except Retry as e:
            logger.debug("Task %s triggered retry", self.task.name)
            return await self.handle_error(e, True, e.when)
        except TaskExpectedError as e:
            logger.warning(
                "Task %s failed with expected error: %s",
                self.task.name,
                self.task.request.id,
            )
            return await self.handle_error(e, True)
        except Exception as e:
            logger.warning(
                "Task %s failed with unexpected error: %s",
                self.task.name,
                self.task.request.id,
                exc_info=True,
            )
            return await self.handle_error(e, self.task.retry_for_error(e))
        else:
            # task executed successfully
            duration = monotonic() - start
            logger.info(
                "Task %s[%s] succeed in %0.4f seconds",
                self.task.name,
                self.task.request.id,
                duration,
            )

            await self.task.on_success(result)

            # TODO: run task callbacks

    async def handle_error(self, error, should_retry, retry_eta=None):
        await self.task.on_failure(
            error,
            self.task.request.id,
            self.task.request.args,
            self.task.request.kwargs,
        )

        # TODO: run task errcallbacks

        if not should_retry:
            return

        retries = self.task.request.retries
        if self.task.max_retries and retries >= self.task.max_retries:
            logger.warning(
                "Task %s[%s] retries exceeded", self.task.name, self.task.request.id
            )
            return

        return retry_eta or self.task.retry_delay()

    async def execute_task(self):
        if self.task.request.time_limit is None:
            return await self.task.run(
                *self.task.request.args, **self.task.request.kwargs
            )

        logger.debug(
            "Executing task %s with %0.4f seconds time limit",
            self.task.name,
            self.task.request.time_limit,
        )

        return await asyncio.wait_for(
            self.task.run(*self.task.request.args, **self.task.request.kwargs),
            timeout=self.task.request.time_limit,
        )

    @property
    def is_expired(self):
        return self.task.request.is_expired()

    @property
    def is_delayed(self):
        return self.task.request.is_delayed

    @property
    def acks_late(self):
        return self.task.acks_late

    async def wait(self):
        secs = self.task.request.countdown()
        if secs:
            await asyncio.sleep(secs)

    def as_retry_message(self, eta=None):
        return self.task.as_message_v2(
            task_id=self.task.request.id,
            name=self.task.name,
            args=self.task.request.args,
            kwargs=self.task.request.kwargs,
            eta=eta or self.task.request.eta,
            expires=self.task.request.expires,
            retries=self.task.request.retries + 1,
            callbacks=self.task.request.callbacks,
            errbacks=self.task.request.errbacks,
            chord=self.task.request.chord,
            chain=self.task.request.chain,
            group=self.task.request.group,
            time_limit=self.task.request.time_limit,
            origin=self.task.request.origin,
            reply_to=self.task.request.reply_to,
        )


def build_tracer(app, task: Task, message: Message, hostname):
    request = Request.from_message(app, message)
    request.hostname = hostname

    exec_task = task.from_request(request)

    return Tracer(exec_task)



import asyncio
import logging

from samping.driver import MockDriver
from samping import App, Rule, ProtocolVersion

logger = logging.getLogger("example")


def driver():
    return MockDriver()


app = App(
    driver_factory=driver,
    default_queue="samping",
)
app.routes = [
    Rule("test_*", "samping"),
    Rule("buggy_*", "buggy"),
]


@app.task(name="test_task")
async def test_task(data: str):
    logger.info("get test_task with data: %s", data)
    await asyncio.sleep(10)


@app.task(name="test_task_v1", protocol_version=ProtocolVersion.V1)
async def test_task_v1(wait: int):
    logger.info("test_task_v1 called with wait: %d", wait)
    await asyncio.sleep(wait)
    return f"waiting completed: {wait}"


@app.task(name="buggy_task")
async def buggy_task(wait: int):
    logger.info("get buggy_task with wait: %d", wait)
    await asyncio.sleep(wait)
    raise RuntimeError("buggy")


@app.tab(name="print_each_minute")
async def print_each_minute():
    logger.info("crontab print each minute executed")
    await asyncio.sleep(2)



import typing
from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from .serialization import loads, dumps
from .exceptions import DecodeError
from .utils.time import maybe_iso8601


class ProtocolVersion(Enum):
    V1 = "v1"
    V2 = "v2"


@dataclass
class Message:
    body: str
    content_type: typing.Optional[str] = None
    content_encoding: typing.Optional[str] = None
    headers: typing.Optional[typing.Dict[str, str]] = None
    properties: typing.Optional[typing.Dict[str, str]] = None
    accept: typing.Optional[typing.List[str]] = None

    def decode(self):
        try:
            return loads(
                self.body, self.content_type, self.content_encoding, self.accept
            )
        except DecodeError:
            # the legacy one use msgpack
            return loads(self.body, "application/x-msgpack", "binary")


@dataclass
class MessageBodyV1:
    id: str
    task: str
    args: typing.Optional[typing.List[typing.Any]] = None
    kwargs: typing.Optional[typing.Dict[str, typing.Any]] = None
    retries: int = 3
    eta: typing.Optional[datetime] = None
    expires: typing.Optional[datetime] = None
    taskset: typing.Optional[str] = None
    chord: typing.Optional[typing.List[typing.Any]] = None
    utc: bool = True
    callbacks: typing.Optional[typing.List[typing.Any]] = None
    errbacks: typing.Optional[typing.List[typing.Any]] = None
    timelimit: typing.Optional[typing.Tuple[float, float]] = None

    @classmethod
    def from_dict(cls, messages):
        eta = messages.get("eta", None)
        expires = messages.get("expires", None)
        if eta and isinstance(eta, str):
            messages["eta"] = maybe_iso8601(eta)
        if expires and isinstance(expires, str):
            messages["expires"] = maybe_iso8601(expires)
        return cls(**messages)

    def encode(self):
        return dumps(
            {
                "id": self.id,
                "task": self.task,
                "args": self.args,
                "kwargs": self.kwargs,
                "retries": self.retries,
                "eta": self.eta.isoformat() if self.eta else None,
                "expires": self.expires.isoformat() if self.expires else None,
            },
            "msgpack",
        )


@dataclass
class MessageBodyV2:
    args: typing.List[typing.Any]
    kwargs: typing.Dict[str, typing.Any]
    embeds: typing.Dict[str, typing.Any]

    def encode(self, serializer):
        return dumps((self.args, self.kwargs, self.embeds), serializer)


TaskMessage = MessageBodyV2



import logging

from ..driver import QueueDriver
from ..messages import Message
from ..types import ASGI3Application
from ..channel import Sender
from ..tasks.tasks import TaskStatus


logger = logging.getLogger("samping")


class QueueCycle:
    def __init__(
        self, message: Message, driver: QueueDriver, sender: Sender[TaskStatus]
    ):
        self.message = message
        self.driver = driver
        self.sender = sender

    def convert_queue_to_protocol(self):
        delivery_info = self.message.properties.get("delivery_info", {})
        headers = {
            "content_type": self.message.content_type,
            "content_encoding": self.message.content_encoding,
        }
        headers.update(self.message.headers or {})
        return {
            "type": "queue",
            "name": delivery_info.get("routing_key", ""),
            "headers": [(k, v) for k, v in headers.items()],
            "properties": [
                (k, v)
                for k, v in self.message.properties.items()
                if k not in ["delivery_info", "delivery_tag"]
            ]
            if self.message.properties
            else [],
            "body": self.message.body,
            "broker": "",
        }

    async def run_app(self, app: ASGI3Application):
        try:
            await self.sender.send(TaskStatus.PENDING)
            scope = self.convert_queue_to_protocol()
            await app(scope, self.receive, self.send)
        except BaseException as exc:
            msg = "Exception in Samping application\n"
            logger.error(msg, exc_info=exc)
        finally:
            await self.sender.send(TaskStatus.FINISHED)

    async def receive(self):
        pass

    async def send(self, message):
        assert message["type"] in ["queue.ack", "queue.nack"]
        if message["type"] == "queue.ack":
            return await self.driver.send_ack(self.message)
        elif message["type.nack"] == "queue.nack":
            return await self.driver.send_nack(self.message, message.get("delay", 0))



def str_to_bytes(s):
    """Convert str to bytes."""
    if isinstance(s, str):
        return s.encode()
    return s


def bytes_to_str(s):
    if isinstance(s, bytes):
        return s.decode(errors="replace")
    return s



import asyncio
import logging
from contextlib import asynccontextmanager

from samping.driver.sqs import SQSDriver
from samping.app import App
from samping.routes import Rule

logger = logging.getLogger("example")


shared_resource = {}

@asynccontextmanager
async def lifespan(app: App):
    logging.getLogger("samping").info("establishing connection to database")
    await asyncio.sleep(10)
    shared_resource["connection"] = "db"
    yield
    shared_resource.clear()


def driver():
    return SQSDriver(
        endpoint_url="http://localhost:9324",
        use_ssl=False,
        prefetch_size=15,
        visibility_timeout=60,
    )


app = App(
    driver_factory=driver,
    default_queue="samping",
    lifespan=lifespan,
)
app.routes = [
    Rule("test_*", "samping"),
    Rule("buggy_*", "buggy"),
]


@app.task(name="test_task")
async def test_task(data: str):
    logger.info("get test_task with data: %s", data)
    await asyncio.sleep(30)


@app.task(name="buggy_task")
async def buggy_task(wait: int):
    logger.info("get buggy_task with wait: %d", wait)
    await asyncio.sleep(wait)
    raise RuntimeError("buggy")


@app.tab(name="print_each_minute")
async def print_each_minute():
    logger.info("crontab print each minute executed")
    await asyncio.sleep(1)



import argparse
import asyncio
import uuid
from .app import test_task, buggy_task


def run_producer(number):
    async def producer():
        await asyncio.gather(
            *[test_task.apply(kwargs={"data": uuid.uuid4().hex}) for number in range(number)]
        )

        await buggy_task.apply(args=[10])

    asyncio.run(producer())


if __name__ == "__main__":
    parser = argparse.ArgumentParser("example producer")

    parser.add_argument("--num-messages", type=int, default=3)

    args = parser.parse_args()

    run_producer(args.num_messages)



import asyncio
import base64
import logging
import typing
import json

from ..protocol import LifeSpanCyle, QueueCycle, beat_cycle
from ..app import App
from ..messages import Message
from ..utils.format import try_to_int
from ..utils.time import maybe_iso8601, utcnow
from ..utils.json import loads
from ..channel import channel, Sender

logger = logging.getLogger("samping")


class LambdaAdapter:
    def __init__(self, app: "App"):
        self.app = app

    def __call__(self, event: dict, context: dict):
        if self.is_queue_event(event):
            messages = self.decode_sqs_messages(event["Records"])
            self.run_in_event_loop(self.run_queue_cycles(messages))
        elif self.is_scheduled_event(event):
            self.run_in_event_loop(self.run_beat_cycle(event))
        else:
            raise TypeError("Invalid SQS event or EventBridge event")
        
    def run_in_event_loop(self, coro):
        loop = asyncio.get_event_loop()
        task = loop.create_task(coro)
        return loop.run_until_complete(task)

    def is_queue_event(self, event):
        return (
            "Records" in event
            and len(event["Records"]) > 0
            and "body" in event["Records"][0]
        )

    def is_scheduled_event(self, event):
        return event.get("source", "") == "aws.events" and "time" in event

    async def run_queue_cycles(self, messages: typing.List[Message]):
        logger.debug("Received %s messages from SQS", len(messages))
        async with LifeSpanCyle(self.app) as lifespan:
            if lifespan.should_exit:
                return
            ev_tx, _ = channel()

            await asyncio.gather(
                *[self.run_queue_cycle(message, ev_tx) for message in messages]
            )

    async def run_beat_cycle(self, event: dict):
        logger.debug("Received EventBridge scheduled event")
        async with LifeSpanCyle(self.app) as lifespan:
            if lifespan.should_exit:
                return
            event_time = maybe_iso8601(event.get("time", ""))
            if not event_time:
                event_time = utcnow()

            await beat_cycle(self.app, event_time)

    async def run_queue_cycle(self, message: Message, sender: Sender):
        queue = QueueCycle(message, self.app.driver, sender)
        await queue.run_app(self.app)

    def decode_sqs_messages(self, messages: typing.List[dict]) -> typing.List[Message]:
        return [self.decode_sqs_message(message) for message in messages]

    def decode_sqs_message(self, message: dict) -> Message:
        body = message.get("body", "")
        attributes = message.get("attributes", {})

        delivery_info = {
            "routing_key": self.get_queue_name(message),
            "approximate_receive_count": try_to_int(
                attributes.get("ApproximateReceiveCount", 1), 1
            ),
        }
        message_attributes = {
            "delivery_info": delivery_info,
            "delivery_tag": message.get("receiptHandle", ""),
        }
        try:
            task_message = loads(body)
        except json.decoder.JSONDecodeError:
            body = base64.standard_b64decode(body)
            return Message(
                body,
                headers={},
                properties=message_attributes,
            )
        else:
            headers = task_message.get("headers", {})
            body = base64.standard_b64decode(task_message.get("body", None))
            properties = task_message.get("properties", {})
            properties.update(message_attributes)
            return Message(
                body,
                headers=headers,
                properties=properties,
                content_type=task_message.get("content_type", None),
                content_encoding=task_message.get("content_encoding", None),
            )

    def get_queue_name(self, message):
        source_names = message.get("eventSourceARN", "").split(":")
        if source_names:
            return source_names[-1]

        return ""



"""Text formatting utilities."""
from __future__ import annotations

import io
import re
from functools import partial
from pprint import pformat
from re import Match
from textwrap import fill
from typing import Any, Callable, Pattern

__all__ = (
    "abbr",
    "abbrtask",
    "dedent",
    "dedent_initial",
    "ensure_newlines",
    "ensure_sep",
    "fill_paragraphs",
    "indent",
    "join",
    "pluralize",
    "pretty",
    "str_to_list",
    "simple_format",
    "truncate",
)

UNKNOWN_SIMPLE_FORMAT_KEY = """
Unknown format %{0} in string {1!r}.
Possible causes: Did you forget to escape the expand sign (use '%%{0!r}'),
or did you escape and the value was expanded twice? (%%N -> %N -> %hostname)?
""".strip()

RE_FORMAT = re.compile(r"%(\w)")


def str_to_list(s: str) -> list[str]:
    """Convert string to list."""
    if isinstance(s, str):
        return s.split(",")
    return s


def dedent_initial(s: str, n: int = 4) -> str:
    """Remove indentation from first line of text."""
    return s[n:] if s[:n] == " " * n else s


def dedent(s: str, sep: str = "\n") -> str:
    """Remove indentation."""
    return sep.join(dedent_initial(l) for l in s.splitlines())


def fill_paragraphs(s: str, width: int, sep: str = "\n") -> str:
    """Fill paragraphs with newlines (or custom separator)."""
    return sep.join(fill(p, width) for p in s.split(sep))


def join(l: list[str], sep: str = "\n") -> str:
    """Concatenate list of strings."""
    return sep.join(v for v in l if v)


def ensure_sep(sep: str, s: str, n: int = 2) -> str:
    """Ensure text s ends in separator sep'."""
    return s + sep * (n - s.count(sep))


ensure_newlines = partial(ensure_sep, "\n")


def abbr(S: str, max: int, ellipsis: str | bool = "...") -> str:
    """Abbreviate word."""
    if S is None:
        return "???"
    if len(S) > max:
        return (
            isinstance(ellipsis, str)
            and (S[: max - len(ellipsis)] + ellipsis)
            or S[:max]
        )
    return S


def abbrtask(S: str, max: int) -> str:
    """Abbreviate task name."""
    if S is None:
        return "???"
    if len(S) > max:
        module, _, cls = S.rpartition(".")
        module = abbr(module, max - len(cls) - 3, False)
        return module + "[.]" + cls
    return S


def indent(t: str, indent: int = 0, sep: str = "\n") -> str:
    """Indent text."""
    return sep.join(" " * indent + p for p in t.split(sep))


def truncate(s: str, maxlen: int = 128, suffix: str = "...") -> str:
    """Truncate text to a maximum number of characters."""
    if maxlen and len(s) >= maxlen:
        return s[:maxlen].rsplit(" ", 1)[0] + suffix
    return s


def pluralize(n: float, text: str, suffix: str = "s") -> str:
    """Pluralize term when n is greater than one."""
    if n != 1:
        return text + suffix
    return text


def pretty(
    value: str, width: int = 80, nl_width: int = 80, sep: str = "\n", **kw: Any
) -> str:
    """Format value for printing to console."""
    if isinstance(value, dict):
        return f"{sep} {pformat(value, 4, nl_width)[1:]}"
    elif isinstance(value, tuple):
        return "{}{}{}".format(
            sep,
            " " * 4,
            pformat(value, width=nl_width, **kw),
        )
    else:
        return pformat(value, width=width, **kw)


def match_case(s: str, other: str) -> str:
    return s.upper() if other.isupper() else s.lower()


def simple_format(
    s: str,
    keys: dict[str, str | Callable],
    pattern: Pattern[str] = RE_FORMAT,
    expand: str = r"\1",
) -> str:
    """Format string, expanding abbreviations in keys'."""
    if s:
        keys.setdefault("%", "%")

        def resolve(match: Match) -> str | Any:
            key = match.expand(expand)
            try:
                resolver = keys[key]
            except KeyError:
                raise ValueError(UNKNOWN_SIMPLE_FORMAT_KEY.format(key, s))
            if callable(resolver):
                return resolver()
            return resolver

        return pattern.sub(resolve, s)
    return s


def remove_repeating_from_task(task_name: str, s: str) -> str:
    """Given task name, remove repeating module names.

    Example:
        >>> remove_repeating_from_task(
        ...     'tasks.add',
        ...     'tasks.add(2, 2), tasks.mul(3), tasks.div(4)')
        'tasks.add(2, 2), mul(3), div(4)'
    """
    # This is used by e.g. repr(chain), to remove repeating module names.
    #  - extract the module part of the task name
    module = str(task_name).rpartition(".")[0] + "."
    return remove_repeating(module, s)


def remove_repeating(substr: str, s: str) -> str:
    """Remove repeating module names from string.

    Arguments:
        task_name (str): Task name (full path including module),
            to use as the basis for removing module names.
        s (str): The string we want to work on.

    Example:

        >>> _shorten_names(
        ...    'x.tasks.add',
        ...    'x.tasks.add(2, 2) | x.tasks.add(4) | x.tasks.mul(8)',
        ... )
        'x.tasks.add(2, 2) | add(4) | mul(8)'
    """
    # find the first occurrence of substr in the string.
    index = s.find(substr)
    if index >= 0:
        return "".join(
            [
                # leave the first occurrence of substr untouched.
                s[: index + len(substr)],
                # strip seen substr from the rest of the string.
                s[index + len(substr) :].replace(substr, ""),
            ]
        )
    return s


StringIO = io.StringIO
_SIO_write = StringIO.write
_SIO_init = StringIO.__init__


class WhateverIO(StringIO):
    """StringIO that takes bytes or str."""

    def __init__(self, v: bytes | str | None = None, *a: Any, **kw: Any) -> None:
        _SIO_init(self, v.decode() if isinstance(v, bytes) else v, *a, **kw)

    def write(self, data: bytes | str) -> int:
        return _SIO_write(self, data.decode() if isinstance(data, bytes) else data)



import asyncio


class Timer:
    def __init__(self, duration: int):
        self.duration = duration
        self._queue = asyncio.Queue()
        self._task = None
        self._stopped = False

    async def chan(self):
        await self._queue.get()

    def stop(self):
        self._stopped = True
        self._task = None

    def start(self):
        if self._task is not None:
            return
        loop = asyncio.get_running_loop()
        task = loop.create_task(self.run_loop())
        # keep reference to task
        self._task = task

    async def run_loop(self):
        while not self._stopped:
            await asyncio.sleep(self.duration)
            await self._queue.put(None)



from .example_app import test_task, test_task_v1
from samping.tasks.request import Request
import pytest


def test_task_as_message():
    t = test_task.to_message(args=["argument-data"], expires=120)
    task_message = t.decode()
    # in v2, we embed list
    assert isinstance(task_message, list)


def test_as_task_v2():
    t = test_task.as_message_v2("id", test_task.name, ["argument-data"])
    assert t.headers.get("retries", 3) == test_task.max_retries
    assert t.headers.get("id") == "id"


def test_create_request_from_message():
    msg = test_task.to_message(args=["argument-data"], expires=120)
    request = Request.from_message(test_task.app, msg)

    assert request.expires is not None
    assert request.args == ["argument-data"]


def test_as_task_v1():
    test = test_task_v1.to_message(args=[30], expires=30)
    assert test.headers.get("id", None) is None


def test_create_request_from_message_v1():
    msg = test_task_v1.to_message(args=[30], expires=30)
    request = Request.from_message(test_task_v1.app, msg)

    assert request.args == [30]
    assert request.expires is not None


@pytest.mark.asyncio
async def test_create_task_executiong_v1():
    msg = test_task_v1.to_message(args=[3], expires=30)
    task = test_task_v1.from_request(Request.from_message(test_task_v1.app, msg))

    assert task.name == test_task_v1.name
    assert task.run == test_task_v1.run
    assert task.request is not None

    assert test_task_v1.request is None

    assert task.app is not None

    result = await task.run(*task.request.args, **task.request.kwargs)

    assert result == "waiting completed: 3"



from collections import UserDict


class Conf(dict):
    def __getattr__(self, k):
        """d.key -> d[key]"""
        try:
            return self[k]
        except KeyError:
            raise AttributeError(f"{type(self).__name__} object has no attribute {k!r}")

    def __setattr__(self, key: str, value) -> None:
        self[key] = value



import asyncio
import typing
import logging

from ..types import ASGI3Application

logger = logging.getLogger("samping")


STATE_TRANSITION_ERROR = "Got invalid state transition on lifespan protocol."


class LifeSpanCyle:
    def __init__(self, app: ASGI3Application):
        self.startup_event = asyncio.Event()
        self.shutdown_event = asyncio.Event()
        self.receive_queue = asyncio.Queue()
        self.error_occured = False
        self.startup_failed = False
        self.shutdown_failed = False
        self.should_exit = False
        self.app = app
        self.state: dict[str, typing.Any] = {}

    async def __aenter__(self):
        logger.info("Waiting for application startup")
        loop = asyncio.get_event_loop()
        main_lifespan = loop.create_task(self.main())

        startup_event = {"type": "lifespan.startup"}
        await self.receive_queue.put(startup_event)
        await self.startup_event.wait()

        if self.startup_failed or self.error_occured:
            logger.error("Application startup failed. Exiting.")
            self.should_exit = True
        else:
            logger.info("Application startup completed.")

        return self

    async def __aexit__(self, exc_type, exc_value, tb):
        if self.error_occured:
            return
        logger.info("Waiting for application shutdown.")
        shutdown_event = {"type": "lifespan.shutdown"}
        await self.receive_queue.put(shutdown_event)
        await self.shutdown_event.wait()

        if self.shutdown_failed or self.error_occured:
            logger.error("Application shutdown failed. Exiting.")
            self.should_exit = True
        else:
            logger.info("Application shutdown complete")

    async def main(self):
        try:
            scope = {
                "type": "lifespan",
                "asgi": {"version": "2.0", "spec_version": "2.0"},
                "state": self.state,
            }
            await self.app(scope, self.receive, self.send)
        except BaseException as exc:
            self.error_occured = True
            if self.startup_failed or self.shutdown_failed:
                return
            else:
                msg = "Exception in 'lifespan' protocol\n"
                logger.error(msg, exc_info=exc)
        finally:
            self.startup_event.set()
            self.shutdown_event.set()

    async def send(self, message: dict):
        assert message["type"] in (
            "lifespan.startup.complete",
            "lifespan.startup.failed",
            "lifespan.shutdown.complete",
            "lifespan.shutdown.failed",
        )

        if message["type"] == "lifespan.startup.complete":
            assert not self.startup_event.is_set(), STATE_TRANSITION_ERROR
            assert not self.shutdown_event.is_set(), STATE_TRANSITION_ERROR
            self.startup_event.set()

        elif message["type"] == "lifespan.startup.failed":
            assert not self.startup_event.is_set(), STATE_TRANSITION_ERROR
            assert not self.shutdown_event.is_set(), STATE_TRANSITION_ERROR
            self.startup_event.set()
            self.startup_failed = True
            if message.get("message"):
                logger.error(message["message"])

        elif message["type"] == "lifespan.shutdown.complete":
            assert self.startup_event.is_set(), STATE_TRANSITION_ERROR
            assert not self.shutdown_event.is_set(), STATE_TRANSITION_ERROR
            self.shutdown_event.set()

        elif message["type"] == "lifespan.shutdown.failed":
            assert self.startup_event.is_set(), STATE_TRANSITION_ERROR
            assert not self.shutdown_event.is_set(), STATE_TRANSITION_ERROR
            self.shutdown_event.set()
            self.shutdown_failed = True
            if message.get("message"):
                logger.error(message["message"])

    async def receive(self):
        return await self.receive_queue.get()



from dataclasses import dataclass
import typing
from collections import namedtuple

T = typing.TypeVar("T")


@dataclass
class Node(typing.Generic[T]):
    count: int = 0
    items: typing.Optional[typing.List[T]] = None
    children: typing.Optional["Node[T]"] = None

    @property
    def leaf(self):
        return self.children is None

    def __len__(self):
        return self.count

    def _update_count(self):
        self.count = len(self.items) if self.items else 0
        if not self.leaf:
            for c in self.children:
                self.count += c.count

    def copy_children(self):
        if not self.children:
            return None
        children = []
        for c in self.children:
            children.append(c.copy())
        return children

    def copy(self):
        return Node(
            count=self.count,
            items=self.items[:] if self.items else None,
            children=self.copy_children(),
        )


PathHint = namedtuple("PathHint", ("used", "path"))


def default_less(a, b):
    return a < b


class Btree(typing.Generic[T]):
    def __init__(
        self, degree, less: typing.Callable[[T, T], bool] = default_less, empty=None
    ):
        self.min, self.max = degree_to_min_max(degree)
        self.less = less
        self.empty = empty
        self.root = None
        self.count = 0

    def _new_node(self, leaf: bool):
        node = Node(items=[])
        if not leaf:
            node.children = []
        return node

    def _binary_search(self, node: Node[T], key: T):
        low, high = 0, len(node.items)
        while low < high:
            h = (low + high) // 2
            if not self.less(key, node.items[h]):
                low = h + 1
            else:
                high = h
        if low > 0 and not self.less(node.items[low - 1], key):
            return low - 1, True

        return low, False

    def set(self, item: T):
        if self.root is None:
            self.root = self._new_node(True)
            self.root.items.append(item)
            self.root.count = 1
            self.count = 1
            return self.empty, False

        prev, replaced, split = self._node_set(self.root, item)
        if split:
            left = self.root.copy()
            right, median = self._node_split(left)
            self.root = self._new_node(False)
            self.root.children.extend([left, right])
            self.root.items.append(median)
            self.root._update_count()
            return self.set(item)
        if replaced:
            return prev, True
        self.count += 1
        return self.empty, False

    def _node_set(self, node: Node[T], item: T):
        i, found = self._binary_search(node, item)
        if found:
            prev = node.items[i]
            node.items[i] = item
            return prev, True, False
        if node.leaf:
            if len(node.items) == self.max:
                return self.empty, False, True
            node.items.append(self.empty)
            copy_at(i + 1, node.items, node.items[i:])
            node.items[i] = item
            node.count += 1
            return self.empty, False, False
        prev, replaced, split = self._node_set(node.children[i], item)
        if split:
            if len(node.items) == self.max:
                return self.empty, False, True
            right, median = self._node_split(node.children[i])
            node.children.append(None)
            copy_at(i + 1, node.children, node.children[i:])
            node.children[i + 1] = right
            node.items.append(self.empty)
            copy_at(i + 1, node.items, node.items[i:])
            node.items[i] = median
            return self._node_set(node, item)
        if not replaced:
            node.count += 1

        return prev, replaced, False

    def _node_split(self, node: Node[T]):
        i = self.max // 2
        median = node.items[i]

        # right node
        right = self._new_node(node.leaf)
        right.items = node.items[i + 1 :]
        if not node.leaf:
            right.children = node.children[i + 1 :]
        right._update_count()

        # left node
        node.items[i] = self.empty
        node.items = node.items[:i]
        if not node.leaf:
            node.children = node.children[: i + 1]
        node._update_count()

        return right, median

    def scan(self):
        yield from self._scan_node(self.root)

    def _scan_node(self, node: Node[T]):
        if node.leaf:
            yield from node.items
            return

        for i, item in enumerate(node.items):
            yield from self._scan_node(node.children[i])

            yield item

        yield from self._scan_node(node.children[len(node.children) - 1])

    def get(self, key: T):
        node = self.root
        while True:
            i, found = self._binary_search(node, key)
            if found:
                return node.items[i], True
            if not node.children:
                return self.empty, False

            node = node.children[i]

    def __len__(self):
        return self.count

    def remove(self, key: T):
        if not self.root:
            return self.empty, False

        prev, deleted = self._remove(self.root, False, key)
        if not deleted:
            return self.empty, False
        if len(self.root.items) == 0 and not self.root.leaf:
            self.root = self.root.children[0]

        self.count -= 1
        if self.count == 0:
            self.root = None
        return prev, True

    def _remove(self, node: Node[T], max: bool, key: T):
        i = 0
        found = False
        if max:
            i, found = len(node.items) - 1, True
        else:
            i, found = self._binary_search(node, key)
        if node.leaf:
            if found:
                # found the items at the leaf, remove it
                prev = node.items[i]
                node.items = node.items[:i] + node.items[i + 1 :]
                node.count -= 1
                return prev, True
            return self.empty, False

        if found:
            if max:
                i += 1
                prev, deleted = self._remove(node.children[i], True, self.empty)
            else:
                prev = node.items[i]
                max_item, _ = self._remove(node.children[i], True, self.empty)
                deleted = True
                node.items[i] = max_item
        else:
            prev, deleted = self._remove(node.children[i], max, key)
        if not deleted:
            return self.empty, False
        node.count -= 1
        if len(node.children[i].items) < self.min:
            self._node_rebalance(node, i)
        return prev, True

    def _node_rebalance(self, node: Node[T], i: int):
        if i == len(node.items):
            i -= 1

        left: Node[T] = node.children[i]
        right: Node[T] = node.children[i + 1]

        if len(left.items) + len(right.items) < self.max:
            # Merges the left and right children nodes together as a single node
            # that includes (left,item,right), and places the contents into the
            # existing left node. Delete the right node altogether and move the
            # following items and child nodes to the left by one slot.

            left.items.append(node.items[i])
            left.items.extend(right.items)
            if not left.leaf:
                left.children.extend(right.children)
            left.count += right.count + 1

            # move the item over one slot
            copy_at(i, node.items, node.items[i + 1 :])
            node.items[len(node.items) - 1] = self.empty
            node.items = node.items[: len(node.items) - 1]

            # move the children over one slot
            copy_at(i + 1, node.children, node.children[i + 2 :])
            node.children[len(node.children) - 1] = None
            node.children = node.children[: len(node.children) - 1]
        elif len(left.items) > len(right.items):
            # move left -> right over one slot
            right.items.append(self.empty)
            copy_at(1, right.items, right.items[:])
            right.items[0] = node.items[i]
            right.count += 1
            node.items[i] = left.items[len(left.items) - 1]
            left.items[len(left.items) - 1] = self.empty
            left.items = left.items[: len(left.items) - 1]
            left.count -= 1

            if not left.leaf:
                right.children.append(None)
                copy_at(1, right.children, right.children[:])
                right.children[0] = left.children[len(left.children) - 1]
                left.children[len(left.children) - 1] = None
                left.children = left.children[: len(left.children) - 1]
                left.count -= right.children[0].count
                right.count += right.children[0].count
        else:
            # move left <- right over one slot
            # Same as above but the other direction
            left.items.append(node.items[i])
            left.count += 1
            node.items[i] = right.items[0]
            copy_at(0, right.items, right.items[1:])
            right.items[len(right.items) - 1] = self.empty
            right.items = right.items[: len(right.items) - 1]
            right.count -= 1

            if not left.leaf:
                left.children.append(right.children[0])
                copy_at(0, right.children, right.children[1:])
                right.children[len(right.childrenn) - 1] = None
                right.children = right.children[: len(right.children) - 1]
                left.count += left.children[len(left.children) - 1].count
                right.count -= left.children[len(left.children) - 1].count

    def ascend(self, pivot: T):
        yield from self._ascend_node(self.root, pivot)

    def _ascend_node(self, node: Node[T], pivot: T):
        i, found = self._binary_search(node, pivot)
        if not found:
            if not node.leaf:
                yield from self._ascend_node(node.children[i], pivot)

        # We are either in case that
        # - node is found, we should iterate throuh it starting at `i`,
        #   the index it was located at.
        # - node is not found, and TODO: fill in.
        while i < len(node.items):
            yield node.items[i]
            if not node.leaf:
                yield from self._scan_node(node.children[i + 1])
            i += 1

    def _node_reverse(self, node: Node[T]):
        if node.leaf:
            yield from reversed(node.items)
            return

        yield from self._node_reverse(node.children[len(node.children) - 1])

        i = len(node.items)
        while i >= 0:
            yield node.items[i]
            yield from self._node_reverse(node.children[i])
            i -= 1

    def descend(self, pivot: T):
        yield from self._descend_node(self.root, pivot)

    def _descend_node(self, node: Node[T], pivot: T):
        i, found = self._binary_search(node, pivot)
        if not found:
            if not node.leaf:
                yield from self._descend_node(node.children[i], pivot)
            i -= 1

        while i >= 0:
            yield node.items[i]
            if not node.leaf:
                yield from self._node_reverse(node.children[i])
            i -= 1


def copy_at(index, dest, source):
    max_copy = min(len(source), len(dest) - index)
    ix = 0
    while ix < max_copy:
        dest[index + ix] = source[ix]
        ix += 1


def degree_to_min_max(deg: int):
    if deg <= 0:
        deg = 32
    elif deg == 1:
        deg = 2
    max = (deg * 2) - 1
    min = max // 2
    return min, max



import base64
import aioboto3
from itertools import zip_longest, count
from datetime import timedelta
from typing import (
    Any,
    AsyncGenerator,
    List,
    Iterable,
    TypeVar,
    Optional,
)
import logging
import asyncio
import json
from botocore.exceptions import ClientError


from ..messages import Message
from . import QueueDriver
from ..utils.iterators import countdown
from ..utils.format import try_to_int
from ..utils.time import (
    get_exponential_backoff_interval,
    maybe_eta_delay_seconds,
    utcnow,
)
from ..utils.json import loads, dumps
from ..utils.encoding import str_to_bytes
from ..utils.btree import Btree

T = TypeVar("T")

message_counter = count()


def less_message(a, b):
    delivery_a = a.properties.get("delivery_info", {})
    delivery_b = b.properties.get("delivery_info", {})

    if delivery_a.get("delivered_at", None) == delivery_b.get("delivered_at", None):
        return delivery_a.get("message_id", 0) < delivery_b.get("message_id", 0)

    return delivery_a.get("delivered_at", None) < delivery_b.get("delivered_at", None)


class Qos:
    def __init__(self, prefetch_count=50, visibility_timeout: int = 300):
        self.prefetch_count = prefetch_count
        self._delivered = Btree(100, less=less_message)
        self.visibility_timeout = visibility_timeout

    def append(self, message):
        self._delivered.set(message)

    def ack(self, message):
        self._delivered.remove(message)

    def can_consume(self):
        pcount = self.prefetch_count
        return not pcount or len(self._delivered) < pcount

    def __len__(self):
        return len(self._delivered)

    def can_consume_max_estimate(self):
        pcount = self.prefetch_count
        return max(pcount - len(self._delivered), 0)

    def get_unacked_messages(self):
        pivot = Message(
            "",
            properties={
                "delivery_info": {
                    "message_id": next(message_counter),
                    "delivered_at": utcnow()
                    - timedelta(seconds=self.visibility_timeout),
                }
            },
        )
        return list(self._delivered.descend(pivot))

    def remove_unacked_messages(self):
        for message in self.get_unacked_messages():
            self._delivered.remove(message)


def chunk(iterable: Iterable[T], n: int) -> Iterable[List[T]]:
    args = [iter(iterable)] * n
    grouper = zip_longest(*args, fillvalue=None)
    for chunk in grouper:
        yield list(filter(None, chunk))


class SQSDriver(QueueDriver):
    def __init__(
        self,
        visibility_timeout: int = 300,
        prefetch_size: int = 10,
        batch_window: int = 10,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        region_name: Optional[str] = None,
        botocore_session: Optional[Any] = None,
        profile_name: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[bool] = None,
        endpoint_url: Optional[str] = None,
    ):
        self.session = aioboto3.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
            botocore_session=botocore_session,
            profile_name=profile_name,
        )
        self.use_ssl = use_ssl
        self.verify = verify
        self.endpoint_url = endpoint_url
        self.visibility_timeout = visibility_timeout
        self.lock = asyncio.Lock()
        self._queues = {}
        self.logger = logging.getLogger("samping")
        self.qos = Qos(
            prefetch_count=prefetch_size, visibility_timeout=visibility_timeout
        )
        self.batch_window = batch_window

    def sqs_client(self):
        return self.session.resource(
            "sqs",
            use_ssl=self.use_ssl,
            verify=self.verify,
            endpoint_url=self.endpoint_url,
        )

    async def get_queue(self, sqs: Any, queue: str):
        async with self.lock:
            if queue in self._queues:
                return await sqs.Queue(self._queues[queue])
            try:
                sqs_queue = await sqs.get_queue_by_name(QueueName=queue)
            except ClientError as error:
                if error.response["Error"]["Code"] in [
                    "AWS.SimpleQueueService.NonExistentQueue",
                    "QueueDoesNotExist",
                ]:
                    # create queue
                    sqs_queue = await sqs.create_queue(
                        QueueName=queue,
                    )
                else:
                    raise error
            self._queues[queue] = sqs_queue.url
            return sqs_queue

    async def send_batch(self, queue: str, messages: List[Message], **kwargs):
        return await self._send_batch(queue, messages)

    async def _send_batch(self, queue: str, messages: List[Message]):
        async with self.sqs_client() as sqs:
            sqs_queue = await self.get_queue(sqs, queue)
            async with BatchWriter(sqs_queue) as batch:
                for id, message in enumerate(messages):
                    await batch.send_message(self.message_to_sqs(message, str(id)))

    async def _fetch_messages_on(self, queues: List[str]):
        async with self.sqs_client() as sqs:
            messages = []
            i = 0
            while i < len(queues):
                sqs_queue = await self.get_queue(sqs, queues[i])
                for max_num in countdown(
                    self.qos.can_consume_max_estimate(),
                    min(self.qos.can_consume_max_estimate(), 10),
                ):
                    current_sqs_messages = await sqs_queue.receive_messages(
                        AttributeNames=["All"],
                        MaxNumberOfMessages=max_num,
                        MessageAttributeNames=["All"],
                        VisibilityTimeout=self.visibility_timeout,
                        WaitTimeSeconds=self.batch_window,
                    )
                    if not current_sqs_messages:
                        break
                    current_messages = await asyncio.gather(
                        *[
                            self._message_to_python(message, queues[i])
                            for message in current_sqs_messages
                        ]
                    )
                    messages.extend(current_messages)
                # break if we already have message, otherwise try next queue
                if not messages:
                    i += 1
                else:
                    break
            return messages

    async def consume(self, queues: str) -> AsyncGenerator[Message, None]:
        queue_names = list(filter(lambda x: x, queues.split(",")))
        retries = 0
        while True:
            next_wait = 0
            if self.qos.can_consume():
                messages = await self._fetch_messages_on(queue_names)
                if messages:
                    self.logger.debug("sqs: received %d messages", len(messages))
                    retries = 0
                    async with self.lock:
                        for message in messages:
                            self.qos.append(message)
                            yield message
                else:
                    retries += 1
                    next_wait = get_exponential_backoff_interval(2, retries, 120)
                    self.logger.info(
                        "sqs: received empty messages, in flight messages: %d",
                        len(self.qos),
                    )
            else:
                retries += 1
                next_wait = get_exponential_backoff_interval(2, retries, 120)
                self.logger.info(
                    "waiting application to process messages, in flight messages: %s",
                    len(self.qos),
                )
                async with self.lock:
                    self.qos.remove_unacked_messages()

            self.logger.debug(
                f"sqs: sleeping for {next_wait} seconds before polling again"
            )
            await asyncio.sleep(next_wait or self.batch_window)

    async def send_ack(self, message: Message):
        ack_id = message.properties.get("delivery_tag", None)
        delivery_info = message.properties.get("delivery_info", {})
        queue = delivery_info.get("routing_key", None)

        if queue and ack_id:
            async with self.sqs_client() as sqs:
                sqs_queue = await self.get_queue(sqs, queue)
                sqs_message = await sqs_queue.Message(ack_id)
                await sqs_message.delete()

            async with self.lock:
                self.qos.ack(message)
        else:
            self.logger.debug(
                "message have empty delivery_tag and delivery_info properties"
            )

    async def send_nack(self, message: Message, delay: int = 1):
        ack_id = message.properties.get("delivery_tag", None)
        delivery_info = message.properties.get("delivery_info", {})
        queue = delivery_info.get("routing_key", None)

        if queue and ack_id:
            async with self.sqs_client() as sqs:
                sqs_queue = await self.get_queue(sqs, queue)
                sqs_message = await sqs_queue.Message(ack_id)
                await sqs_message.change_visibility(VisibilityTimeout=delay)

            async with self.lock:
                self.qos.ack(message)
        else:
            self.logger.debug(
                "message have empty routing_key and delivery_info properties"
            )

    async def _message_to_python(self, sqs_message, queue: str) -> Message:
        bodyStr = await sqs_message.body
        attributes = await sqs_message.attributes

        delivery_info = {
            "routing_key": queue,  # set the queue as routing key
            "approximate_receive_count": try_to_int(
                attributes.get("ApproximateReceiveCount", 1), 1
            ),
            "message_id": next(message_counter),
            "delivered_at": utcnow(),
        }
        message_attributes = {
            "delivery_info": delivery_info,
            "delivery_tag": sqs_message.receipt_handle,
        }
        try:
            message = loads(bodyStr)
        except json.decoder.JSONDecodeError:
            body = base64.standard_b64decode(bodyStr)
            return Message(
                body,
                headers={},
                properties=message_attributes,
            )
        else:
            headers = message.get("headers", {})
            body = base64.standard_b64decode(message.get("body", None))
            properties = message.get("properties", {})
            properties.update(message_attributes)
            return Message(
                body,
                headers=headers,
                properties=properties,
                content_type=message.get("content_type", None),
                content_encoding=message.get("content_encoding", None),
            )

    def _to_delivery_dict(self, message: Message):
        body = base64.standard_b64encode(str_to_bytes(message.body)).decode()
        properties = message.properties or {}
        if "delivery_info" in properties:
            del properties["delivery_info"]
        if "delivery_tag" in properties:
            del properties["delivery_tag"]

        return {
            "headers": message.headers,
            "properties": properties,
            "content_type": message.content_type,
            "content_encoding": message.content_encoding,
            "body": body,
        }

    def message_to_sqs(self, message: Message, message_id=None):
        sqs_message = {
            "MessageBody": dumps(self._to_delivery_dict(message)),
        }
        if message_id:
            sqs_message["Id"] = message_id

        delay = maybe_eta_delay_seconds(message.headers.get("eta", None))
        if delay:
            # maximum delay for sqs is 15 minute
            sqs_message["DelaySeconds"] = max(0, min(delay, 900))

        return sqs_message

    async def size(self, queue: str) -> int:
        async with self.sqs_client() as sqs:
            sqs_queue = await self.get_queue(sqs, queue)
            attributes = await sqs_queue.attributes
            return attributes.get("ApproximateNumberOfMessages", 0)


class BatchWriter:
    def __init__(self, queue, flush_amount=10, flush_size=250000):
        self._queue = queue
        self._messages = []
        self._current_size = 0
        self._flush_size = flush_size
        self._flush_amount = flush_amount
        self._lock = asyncio.Lock()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._messages:
            await self._flush()

    def get_size(self, message):
        body = message.get("MessageBody", "").encode("utf-8")
        return len(body)

    async def send_message(self, message):
        async with self._lock:
            current_size = self._current_size + self.get_size(message)
            if current_size > self._flush_size:
                await self._flush()
                self._messages.append(message)
                self._current_size += self.get_size(message)
                if len(self._messages) >= self._flush_amount:
                    await self._flush()
            else:
                self._current_size = current_size
                self._messages.append(message)
                if len(self._messages) >= self._flush_amount:
                    await self._flush()

    async def _flush(self):
        response = await self._queue.send_messages(Entries=self._messages)
        self._messages = []
        if len(response.get("Failed", [])) > 0:
            first = response["Failed"][0]
            raise Exception(
                "sqs send_batch: failed for %d message(s): %s"
                % (len(response["Failed"]), first.get("Message", ""))
            )
        self._current_size = 0



import typing
from datetime import datetime
from dataclasses import dataclass
from ..messages import Message, MessageBodyV1
from ..utils.format import try_to_int
from ..utils.time import maybe_iso8601


@dataclass
class Request:
    # any for now, it reference to samping app
    app: typing.Any

    # the unique id of the executing task
    id: str

    # Custom ID used for things like de-duplication. Usually the same as `id`.
    correlation_id: str

    # args parameters used to call this task
    args: typing.List[typing.Any]

    # keyword arguements parameter used to call this task
    kwargs: typing.MutableMapping[str, typing.Any]

    # How many times the current task has been retries
    retries: int = 0

    # the unique id of the task's group, if this task is a member.
    group: typing.Optional[str] = None

    # Name of the host that sent this task
    origin: typing.Optional[str] = None

    # The original ETA of the task
    eta: typing.Optional[datetime] = None

    # The original expiration of the task
    expires: typing.Optional[datetime] = None

    # Node name of the worker instance executing the task
    hostname: typing.Optional[str] = None

    # Where to send reply to (queue name)
    reply_to: typing.Optional[str] = None

    # time limit
    time_limit: typing.Optional[int] = None

    # A list of signatures to call if the task exited successfully.
    callbacks: typing.Optional[typing.List[typing.Any]] = None

    # errbacks
    errbacks: typing.Optional[typing.List[typing.Any]] = None

    # A list of signatures that only executes after all of the tasks in a group has finished executing.
    chord: typing.Optional[typing.List[typing.Any]] = None

    # chain
    chain: typing.Optional[typing.List[typing.Any]] = None

    @classmethod
    def from_message(cls, app, message: Message):
        if message.headers.get("id", None):
            return cls.from_message_v2(app, message)
        return cls.from_message_v1(app, message)

    @classmethod
    def from_message_v2(cls, app, message: Message):
        args, kwargs, embeds = message.decode()

        eta = maybe_iso8601(message.headers.get("eta", None))
        expires = maybe_iso8601(message.headers.get("expires", None))

        return Request(
            app=app,
            id=message.headers.get("id", None),
            correlation_id=message.properties.get("correlation_id", None),
            args=args,
            kwargs=kwargs,
            retries=try_to_int(message.headers.get("retries", 0)),
            eta=eta,
            expires=expires,
            group=message.headers.get("group", None),
            origin=message.headers.get("origin", None),
            reply_to=message.properties.get("reply_to", None),
            time_limit=get_time_limit(message.headers.get("time_limit", None)),
            chord=embeds.get("chord", None),
            callbacks=embeds.get("callbacks", None),
            errbacks=embeds.get("errbacks", None),
            chain=embeds.get("chain", None),
        )

    @classmethod
    def from_message_v1(cls, app, message: Message):
        """All information for v1 is in message body"""
        task_message = MessageBodyV1.from_dict(message.decode())

        return Request(
            app=app,
            id=task_message.id,
            correlation_id=task_message.id,
            args=task_message.args,
            kwargs=task_message.kwargs,
            retries=task_message.retries,
            eta=task_message.eta,
            expires=task_message.expires,
            group=task_message.taskset,
            time_limit=get_time_limit(task_message.timelimit),
            chord=task_message.chord,
            callbacks=task_message.callbacks,
            errbacks=task_message.errbacks,
        )

    @property
    def is_delayed(self):
        """Check if the request has a future ETA"""
        return self.eta is not None

    def countdown(self, now=None):
        """Get the TTL in seconds if the task has a future ETA."""
        if not self.eta:
            return None
        now = now or self.app.now
        countdown = (self.eta - now()).total_seconds()
        return None if countdown < 0 else countdown

    def is_expired(self, now=None):
        """Check if the request is expired"""
        if self.expires is None:
            return False
        now = now or self.app.now
        return (now() - self.expires).total_seconds() >= 0


def get_time_limit(time_limit):
    if not time_limit:
        return time_limit

    if isinstance(time_limit, (int, float)):
        return time_limit

    soft_limit, hard_limit = time_limit

    if soft_limit is not None and hard_limit is not None:
        return min(soft_limit, hard_limit)
    if soft_limit is None:
        return soft_limit
    if hard_limit is None:
        return hard_limit



import typing

AppType = typing.TypeVar("AppType")

Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]

Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]

StatelessLifespan = typing.Callable[[AppType], typing.AsyncContextManager[None]]
StatefulLifespan = typing.Callable[
    [AppType], typing.AsyncContextManager[typing.Mapping[str, typing.Any]]
]
Lifespan = typing.Union[StatelessLifespan[AppType], StatefulLifespan[AppType]]

ASGI3Application = typing.Callable[
    [
        Scope,
        Receive,
        Send,
    ],
    typing.Awaitable[None],
]



import asyncio
from typing import TypeVar, Generic, Tuple

T = TypeVar("T")


class Sender(Generic[T]):
    def __init__(self, queue: asyncio.Queue[T]):
        self._queue = queue

    async def send(self, value: T):
        await self._queue.put(value)


class Receiver(Generic[T]):
    def __init__(self, queue: asyncio.Queue[T]):
        self._queue = queue

    async def recv(self) -> T:
        value = await self._queue.get()
        self._queue.task_done()

        return value

    async def with_recv(self, action):
        value = await self._queue.get()
        await action(value)
        self._queue.task_done()


def channel(maxsize=0) -> Tuple[Sender[T], Receiver[T]]:
    queue = asyncio.Queue(maxsize=maxsize)

    return Sender(queue), Receiver(queue)



from typing import Dict, List, TypedDict, Any, Union
from datetime import datetime, timedelta
from enum import Enum
from collections.abc import Mapping
import uuid
import numbers

from ..exceptions import Retry
from ..utils.time import maybe_make_aware, get_exponential_backoff_interval
from ..messages import Message, MessageBodyV1, MessageBodyV2, ProtocolVersion
from .request import Request

#: earliest date supported by time.mktime.
INT_MIN = -2147483648


class TaskInfo(TypedDict):
    args: List[Any]
    kwargs: Dict[str, Any]
    task_id: Union[None, str]
    queue: Union[None, str]
    eta: Union[None, datetime]
    expires: Union[None, datetime]


class class_property:
    def __init__(self, getter=None, setter=None):
        if getter is not None and not isinstance(getter, classmethod):
            getter = classmethod(getter)
        if setter is not None and not isinstance(setter, classmethod):
            setter = classmethod(setter)
        self.__get = getter
        self.__set = setter

        info = getter.__get__(object)  # just need the info attrs.
        self.__doc__ = info.__doc__
        self.__name__ = info.__name__
        self.__module__ = info.__module__

    def __get__(self, obj, type=None):
        if obj and type is None:
            type = obj.__class__
        return self.__get.__get__(obj, type)()

    def __set__(self, obj, value):
        if obj is None:
            return self
        return self.__set.__get__(obj)(value)

    def setter(self, setter):
        return self.__class__(self.__get, setter)


class Task:
    _app = None

    _request = None

    name = None

    max_retries = 3

    serializer = None

    # the time limit of this task allowed in seconds
    time_limit = 300

    # When enabled message for this will be acknowledged **after**
    # the task has been executed, and not *right before* (the default behaviour)
    acks_late = None

    # default strategy to calculate retry delay eta
    # this variable set the minimum retry delay allowed
    min_retry_delay = None

    # default strategy to calculate delay eta
    # this variable set the maximum retry delay allowed
    max_retry_delay = None

    protocol_version = ProtocolVersion.V2

    __bound__ = False

    def retry(self, exc=None, when=None, **kwargs):
        raise Retry(exc=exc, when=when, **kwargs)

    @classmethod
    def bind(cls, app):
        cls.__bound__ = True
        cls._app = app

        cls.on_bound(app)

        return app

    @classmethod
    def on_bound(cls, app):
        pass

    @classmethod
    def _get_app(cls):
        if not cls.__bound__:
            # The app property's __set__  method is not called
            # if Task.app is set (on the class), so must bind on use.
            cls.bind(cls._app)
        return cls._app

    app = class_property(_get_app, bind)

    @property
    def request(self) -> Request | None:
        return self._request

    def from_request(self, request: Request) -> "Task":
        cls = self.__class__
        task = cls()
        task._request = request

        return task

    async def run(self, *args, **kwargs):
        raise NotImplementedError("Tasks must define the run method.")

    async def on_success(self, result=None):
        pass

    async def on_failure(self, exc, task_id, args, kwargs):
        pass

    def retry_for_error(self, exc) -> bool:
        """Called when an error raised during execution of task. Return
        True if we wish to retry
        """
        return True

    def retry_delay(self, now=None):
        """Called when this task failed, return time the task should be retried"""
        delay_secs = get_exponential_backoff_interval(
            2, self.request.retries, self.get_max_retry_delay()
        )
        return max(delay_secs, self.get_min_retry_delay())

    def get_min_retry_delay(self):
        if self.min_retry_delay:
            return self.min_retry_delay

        return 0

    def get_max_retry_delay(self):
        if self.max_retry_delay:
            return self.max_retry_delay

        return 3600

    def to_message(
        self,
        args=None,
        kwargs=None,
        task_id=None,
        countdown=None,
        eta=None,
        expires=None,
        retries=None,
        timezone=None,
        **options,
    ) -> Message:
        if not task_id:
            task_id = str(uuid.uuid4())

        as_message = (
            self.as_message_v2
            if self.protocol_version == ProtocolVersion.V2
            else self.as_message_v1
        )

        return as_message(
            task_id=task_id,
            name=self.name,
            args=args,
            countdown=countdown,
            kwargs=kwargs,
            eta=eta,
            expires=expires,
            timezone=timezone,
            retries=retries or 0,
            **options,
        )

    def as_message_v2(
        self,
        task_id,
        name,
        args=None,
        kwargs=None,
        countdown=None,
        eta=None,
        expires=None,
        timezone=None,
        **options,
    ) -> Message:
        args = args or ()
        kwargs = kwargs or {}

        if not isinstance(args, (list, tuple)):
            raise TypeError("task args must be a list or tuple")
        if not isinstance(kwargs, Mapping):
            raise TypeError("task keyword arguments must be a mapping")

        if task_id is None:
            task_id = str(uuid.uuid4())

        if countdown:
            self._verify_seconds(countdown, "countdown")
            now = self.app.now()
            timezone = timezone or self.app.timezone
            eta = maybe_make_aware(now + timedelta(seconds=countdown), tz=timezone)

        if isinstance(expires, numbers.Real):
            self._verify_seconds(expires, "expires")
            timezone = timezone or self.app.timezone
            expires = maybe_make_aware(
                self.app.now() + timedelta(seconds=expires), tz=timezone
            )

        if not isinstance(eta, str):
            eta = eta and eta.isoformat()

        root_id = options.get("root_id", None)
        if not root_id:
            root_id = task_id

        reply_to = options.get("reply_to", "")

        headers = {
            "lang": "py",
            "task": name,
            "id": task_id,
            "eta": eta,
            "expires": expires,
            "group": options.get("group", None),
            "group_id": options.get("group_id", None),
            "retries": options.get("retries", 3),
            "timelimit": [options.get("time_limit", self.time_limit), None],
            "root_id": root_id,
            "parent_id": options.get("parent_id", None),
            "origin": options.get("origin_id", None),
            "ignore_result": options.get("ignore_result", False),
            "replaced_task_nesting": options.get("replaced_task_nesting", 0),
        }

        body = MessageBodyV2(
            args=args,
            kwargs=kwargs,
            embeds={
                "callbacks": options.get("callbacks", None),
                "errbacks": options.get("errbacks", None),
                "chain": options.get("chain", None),
                "chord": options.get("chord"),
            },
        )

        content_type, content_encoding, data = body.encode(self.serializer)

        return Message(
            headers=headers,
            content_type=content_type,
            content_encoding=content_encoding,
            properties={
                "correlation_id": task_id,
                "reply_to": reply_to,
            },
            body=data,
        )

    def as_message_v1(
        self,
        task_id,
        name,
        args=None,
        kwargs=None,
        countdown=None,
        eta=None,
        expires=None,
        timezone=None,
        **options,
    ):
        args = args or ()
        kwargs = kwargs or {}

        if not isinstance(args, (list, tuple)):
            raise TypeError("task args must be a list or tuple")
        if not isinstance(kwargs, Mapping):
            raise TypeError("task keyword arguments must be a mapping")

        if task_id is None:
            task_id = str(uuid.uuid4())

        if countdown:
            self._verify_seconds(countdown, "countdown")
            now = self.app.now()
            timezone = timezone or self.app.timezone
            eta = maybe_make_aware(now + timedelta(seconds=countdown), tz=timezone)

        if isinstance(expires, numbers.Real):
            self._verify_seconds(expires, "expires")
            timezone = timezone or self.app.timezone
            expires = maybe_make_aware(
                self.app.now() + timedelta(seconds=expires), tz=timezone
            )

        if not isinstance(eta, str):
            eta = eta and eta.isoformat()

        body = MessageBodyV1(
            id=task_id,
            task=name,
            args=args,
            kwargs=kwargs,
            retries=options.get("retries", 0),
            eta=eta,
            expires=expires,
            timelimit=options.get("time_limit", self.time_limit),
        )
        content_type, content_encoding, data = body.encode()

        return Message(
            headers={},
            content_type=content_type,
            content_encoding=content_encoding,
            body=data,
        )

    async def batch(self, batches: List[TaskInfo], queue: str = None):
        messages = [self.to_message(**batch) for batch in batches]
        return await self._app.send_batch(
            messages, queue=queue or self._app.task_route(self)
        )

    async def apply(
        self,
        args=None,
        kwargs=None,
        task_id: str = None,
        queue=None,
        countdown=None,
        eta=None,
        expires=None,
        **opts,
    ):
        queue = queue or self._app.task_route(self)
        return await self._app._send_message(
            self.to_message(
                args=args,
                kwargs=kwargs,
                task_id=task_id,
                eta=eta,
                expires=expires,
                countdown=countdown,
                **opts,
            ),
            queue=queue,
        )

    def _verify_seconds(self, s, what):
        if s < INT_MIN:
            raise ValueError(f"{what} is out of range: {s!r}")
        return s


class CronJob:
    name = None

    expression = "* * * * *"

    async def run(self):
        raise NotImplementedError("CronJob must define the run method.")


class TaskStatus(Enum):
    PENDING = 1
    FINISHED = 2



from datetime import timezone, datetime
import pytest

from .example_app import app, test_task, test_task_v1, buggy_task


async def send(message):
    pass


async def receive():
    pass


def test_config_setting():
    tz = app.timezone
    assert isinstance(tz, timezone)


def test_timezone_config():
    app.conf.timezone = "Asia/Jakarta"
    assert app.conf.timezone == "Asia/Jakarta"

    now = app.now()
    assert not now.utcoffset()


@pytest.mark.asyncio
async def test_send_beat_event():
    event = {
        "type": "beat",
        "time": datetime.now(timezone.utc).isoformat(),
    }

    await app(event, receive, send)


@pytest.mark.asyncio
async def test_send_queue_event():
    msg = test_task.to_message(args=["argument-data"], expires=120, countdown=20)
    msg.headers["content_type"] = msg.content_type
    msg.headers["content_encoding"] = msg.content_encoding

    event = {
        "type": "queue",
        "name": "samping",
        "headers": msg.headers,
        "properties": msg.properties,
        "hostname": "",
        "body": msg.body,
    }

    await app(event, receive, send)


@pytest.mark.asyncio
async def test_send_queue_event_v1():
    msg = test_task_v1.to_message(args=[5], expires=120)
    msg.headers["content_type"] = msg.content_type
    msg.headers["content_encoding"] = msg.content_encoding
    event = {
        "type": "queue",
        "name": "samping",
        "headers": msg.headers,
        "properties": msg.properties,
        "hostname": "",
        "body": msg.body,
    }

    await app(event, receive, send)


@pytest.mark.asyncio
async def test_executing_buggy_task():
    msg = buggy_task.to_message(args=[5])
    msg.headers["content_type"] = msg.content_type
    msg.headers["content_encoding"] = msg.content_encoding
    event = {
        "type": "queue",
        "name": "samping",
        "headers": msg.headers,
        "properties": msg.properties,
        "hostname": "",
        "body": msg.body,
    }

    await app(event, receive, send)



import datetime

from samping.messages import Message, MessageBodyV1, MessageBodyV2
from samping.serialization import dumps
from samping.utils.time import utcnow
from samping.utils.compat import msgpack_dumps


def test_encode_task_message():
    task = {"id": 9023, "expires": utcnow()}
    content_type, content_encoding, data = dumps(task)

    message = Message(
        data,
        content_type=content_type,
        content_encoding=content_encoding,
        headers={},
    )
    decoded = message.decode()

    assert task.get("id") == decoded.get("id")


def test_encode_task_legacy_message():
    task = {"id": 9023, "expires": utcnow()}
    body = msgpack_dumps(task)

    message = Message(body)
    decoded = message.decode()

    assert task.get("id") == decoded.get("id")


def test_encode_task_messagev1_with_eta():
    task_message = MessageBodyV1(
        id="id", task="test_encode", eta=datetime.datetime.now()
    )
    content_type, content_encoding, body = task_message.encode()
    assert isinstance(body, bytes)
    message = Message(
        body=body, content_type=content_type, content_encoding=content_encoding
    )
    task_message_2 = MessageBodyV1.from_dict(message.decode())

    assert isinstance(task_message_2.eta, datetime.datetime)


def test_encode_task_message_v2():
    task_message = MessageBodyV2(
        ["data"],
        {},
        {"callbacks": None, "errbacks": None, "chain": None, "chord": None},
    )
    content_type, encoding, data = task_message.encode("json")

    assert content_type == "application/json"
    assert encoding == "utf-8"
    assert isinstance(data, str)



from typing import List, Protocol, AsyncGenerator
from abc import abstractmethod
from collections import OrderedDict
import asyncio

from ..messages import Message


class QueueDriver(Protocol):
    @abstractmethod
    async def send_batch(self, queue: str, messages: List[Message], **kwargs) -> None:
        ...

    @abstractmethod
    async def consume(self, queues: str) -> AsyncGenerator[Message, None]:
        ...

    @abstractmethod
    async def send_ack(self, message: Message) -> None:
        ...

    @abstractmethod
    async def send_nack(self, message: Message, delay: int = 1) -> None:
        ...


class MockDriver(Protocol):
    def __init__(self):
        self._lock = asyncio.Lock()
        self.queues = OrderedDict()

    async def send_batch(self, queue, message: List[Message], **kwargs):
        async with self._lock:
            if queue not in self.queues:
                self.queues[queue] = message
            else:
                messages = self.queues[queue]
                messages.extend(queue)

    async def send_concume(self, queues):
        raise NotImplemented("not implemented")

    async def send_ack(self, message):
        pass

    async def send_nack(self, message, delay):
        pass



from samping.utils.btree import Btree


def test_btree_set():
    tree = Btree(10)

    keys = list(range(1000))

    for key in keys:
        _, replaced = tree.set(key)
        assert not replaced

    assert len(tree) == len(keys)


def test_btree_get_remove():
    tree = Btree(10)

    keys = list(range(500))
    for key in keys:
        _, replaced = tree.set(key)
        assert not replaced

    val, ok = tree.get(40)
    assert ok
    assert val == 40

    val, ok = tree.remove(40)
    assert ok
    assert val == 40

    val, ok = tree.get(40)
    assert not ok
    assert val == tree.empty



from datetime import timedelta

from samping.driver.sqs import Qos
from samping.messages import Message
from samping.utils.time import utcnow


def test_qos_get_unacked_messages():
    qos = Qos(50, visibility_timeout=300)

    for i in range(10):
        message = Message(
            "",
            properties={
                "delivery_info": {
                    "message_id": i,
                    "delivered_at": utcnow() - timedelta(seconds=320 + i),
                }
            },
        )
        qos.append(message)

    for i in range(11, 20):
        message = Message(
            "",
            properties={
                "delivery_info": {
                    "message_id": i,
                    "delivered_at": utcnow() - timedelta(seconds=i),
                }
            },
        )
        qos.append(message)

    assert qos.can_consume()

    # this messages need to be removed
    messages = qos.get_unacked_messages()
    assert len(messages) == len(list(range(10)))

    qos.remove_unacked_messages()

    assert len(qos) == len(list(range(11, 20)))



from datetime import datetime, date
import uuid
import decimal
import msgpack


def encode_non_standard_msgpack(obj):
    if isinstance(obj, (decimal.Decimal, uuid.UUID)):
        return str(obj)
    if isinstance(obj, (datetime, date)):
        if not isinstance(obj, datetime):
            obj = datetime(obj.year, obj.month, obj.day, 0, 0, 0, 0)
        r = obj.isoformat()
        if r.endswith("+00:00"):
            r = r[:-6] + "Z"
        return r
    return obj


def msgpack_dumps(s, **kwargs):
    return msgpack.packb(s, default=encode_non_standard_msgpack, use_bin_type=True)



from .app import App
from .exceptions import Retry
from .routes import Rule
from .messages import ProtocolVersion
from .worker import Worker, run_worker
from .adapters.aws_lambda import LambdaAdapter


__all__ = [
    "App",
    "Retry",
    "Rule",
    "ProtocolVersion",
    "Worker",
    "run_worker",
    "LambdaAdapter",
]
